/*
 * Hanashizume no Matsuri for Palm
 * Copyright (c) 2001 TF <tf@denpa.org>
 * All rights reserverd.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *      This product includes software developed by TF.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: pdbtar.c,v 1.5 2001/02/19 17:04:31 tf Exp $
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <time.h>

/* Exact value of "Jan 1, 1970 0:00:00 GMT" - "Jan 1, 1904 0:00:00 GMT" */
#define PILOT_TIME_DELTA (unsigned)(2082844800)
#define PI_HDR_SIZE 78
#define PALMDB_NAME_LEN 32
#define NAME_LEN 256

#define TABLE_RECORD_SIZE 2

static unsigned long unix_time_to_pilot_time(time_t);
void parse_text_filename(const char *, int *, int *);
void parse_cg_filename(const char *, int *);

void put2b(int, FILE *);
void put3b(int, FILE *);
void put4b(int, FILE *);

static int get_uid_from_filename(const char *);

typedef struct {
  char name[NAME_LEN];
  int size;
} fileinfo;

typedef struct {
  unsigned short len; /* 2 bytes */
} table_record;

/*
 * header:
 * 32           name
 * 2            flags
 * 2            version
 * 4            creation time
 * 4            modification time
 * 4            backup time
 * 4            modification number
 * 4            app info offset
 * 4            sort info offset
 * 4            type
 * 4            creator
 * 4            unique id seed (I think it is just garbage)
 * 4            next record list id (normally 0, or ptr to extended hdr)
 * 2            num records for this header
 */

int uid;

void
output_header(FILE *fp, const char *name, const char *id, int num)
{
  int i, len;

  /* name */
  len = strlen(name);
  if (len > 32) {
    fprintf(stderr, "the name length should be less than 32 chars.\n");
    exit(1);
  }
  
  for (i = 0; i < len; i++) {
    fputc(name[i], fp);
  }
  for (; i < PALMDB_NAME_LEN; i++) {
    fputc('\0', fp);
  }

  /* <FLAGS>
   * 0x0001: Resource DB, instead of record DB
   * 0x0002: DB is read only
   * 0x0004: AppInfo data has been modified
   * 0x0008: DB is tagged for generic backup
   * 0x8000: DB is currently open
   * 0x0010: Newer version may be installed over open DB (v2.0 specific)
   * 0x0020: Reset after installation (v2.0 specific)
   */
  put2b(0x0000, fp);

  /* version */
  put2b(0, fp);

  /* ctime */
  put4b(unix_time_to_pilot_time(time(NULL)), fp);

  /* mtime */
  put4b(unix_time_to_pilot_time(time(NULL)), fp);

  /* backup time */
  put4b(0, fp);

  /* modification number */
  put4b(0, fp);

  /* app info offset */
  put4b(0, fp);

  /* sort info offset */
  put4b(0, fp);

  /* type */
  put4b(0x44415441, fp); /* 0x44415441 means 'DATA' database type. */

  /* creator */
  for (i = 0; i < 4; i++) {
    fputc(id[i], fp);
  }

  /* unique id seed (this may be just garbage) */
  put4b(0, fp);
  
  /* next record list id (normally 0, or ptr to extended hdr) */
  put4b(0, fp);
  
  /* number of record */
  put2b(num, fp);
}

static unsigned long
unix_time_to_pilot_time(time_t t)
{
  return (unsigned long)((unsigned long)t + PILOT_TIME_DELTA);
}

void
put4b(int val, FILE *fp)
{
  unsigned char b1, b2, b3, b4;

  b1 = val & 0xff;
  b2 = (val >> 8) & 0xff;
  b3 = (val >> 16) & 0xff;
  b4 = (val >> 24) & 0xff;

  fputc(b4, fp);
  fputc(b3, fp);
  fputc(b2, fp);
  fputc(b1, fp);
}

void
put3b(int val, FILE *fp)
{
  unsigned char b1, b2, b3;

  b1 = val & 0xff;
  b2 = (val >> 8) & 0xff;
  b3 = (val >> 16) & 0xff;

  fputc(b3, fp);
  fputc(b2, fp);
  fputc(b1, fp);
}

void
put2b(int val, FILE *fp)
{
  unsigned char b1, b2;

  b1 = val & 0xff;
  b2 = (val >> 8) & 0xff;

  fputc(b2, fp);
  fputc(b1, fp);
}

void
output_record_entry_header(FILE *fp, fileinfo *fi, int num)
{
  int i;
  int pos;

  pos = 0x50 + 8 * (num + 1);

  /* the first record is a table including header info. */
  /* position */
  put4b(pos, fp);
    
  /* flag  */
  fputc(0, fp);
    
  /* uid */
  put3b(0, fp);
  pos += num * TABLE_RECORD_SIZE;
  
  for (i = 0; i < num; i++) {
    /* position */
    put4b(pos, fp);
    
    /* flag  */
    fputc(0, fp);
    
    /* uid */
    put3b(get_uid_from_filename(fi[i].name), fp);
    pos += fi[i].size;
  }
};

void
output_delimiter(FILE *fp)
{
  fputc('\0', fp);
  fputc('\0', fp);
}

void
output_app_info()
{
  return;
}

void
output_sort_info()
{
  return;
}

void
output_record(FILE *outfp, fileinfo *fi, int num)
{
  int i;
  FILE *infp;
  unsigned char *buf;
  
  for (i = 0; i < num; i++) {
    if (fi[i].size == 0) {
      continue;
    }
    fprintf(stderr, "file: %s\n", fi[i].name);
    infp = fopen(fi[i].name, "rb");
    buf = (unsigned char *)calloc(fi[i].size, sizeof(unsigned char));
    fread(buf, 1, fi[i].size, infp);
    fwrite(buf, 1, fi[i].size, outfp);
    free(buf);
    fclose(infp);
  }
}     


fileinfo
*collect_file_info(int num, char *argv[])
{
  fileinfo *res;
  int i, fd;
  struct stat sb;

  res = (fileinfo *)calloc(num, sizeof(fileinfo));

  for (i = 0; i < num; i++) {
#ifndef __NO_UNIX
    if ((fd = open(argv[i], 0, O_RDONLY)) < 0) {
      fprintf(stderr, "not found %s\n", argv[i]);
      res[i].size = 0;
      continue;
    }
    if (fstat(fd, &sb) < 0) {
      perror("fstat");
      exit(1);
    }
    fprintf(stderr, "%s: %d\n", argv[i], (int)sb.st_size);
    strncpy(res[i].name, argv[i], NAME_LEN);
    res[i].size = sb.st_size;
    close(fd);
#else
    {
      FILE *fp;

      fp = fopen(argv[i], "rb");
      if (fp == NULL) {
        fprintf(stderr, "not found %s\n", argv[i]);
	res[i].size = 0;
        continue;
      }
      fseek(fp, 0L, SEEK_END); 
      strncpy(res[i].name, argv[i], NAME_LEN);
      res[i].size = ftell(fp);
      fprintf(stderr, "%s: %d\n", argv[i], res[i].size);
      fclose(fp);
    }
#endif
  }
  
  return res;
}

static void
output_first_record(FILE *fp, fileinfo *fi, int num)
{
  int i;

  for (i = 0; i < num; i++) {
//    put2b(0, fp); /* arg */
    put2b(fi[i].size, fp); /* size */
  }
}

static int
get_uid_from_filename(const char *filename)
{
  char *p;
  int block, chunk, id;

  p = strrchr(filename, '/');
  if (p == NULL) {
    p = filename;
  } else {
    p++;
  }
  
  if (!strncmp(p, "0x", 2)) {
    parse_text_filename(p, &block, &chunk);
    id = 2500 + block * 20 + chunk;
    fprintf(stderr, "%s: %d(%d:%d)\n", p, id, block, chunk);
  } else if (!strncmp(p, "CG", 2) || !strncmp(p, "HNTI", 4)) {
    parse_cg_filename(p, &id);
    fprintf(stderr, "%s: %d\n", p, id);
  } else {
    id = uid++;
  }
  
  return id;
}

void
parse_text_filename(const char *name, int *block, int *chunk)
{
  char *upper, *lower, *p;
  
  upper = strdup(name);
  
  for (p = upper; *p != '_'; p++);
  *p = '\0';
  *block = strtol(upper, NULL, 16);
  
  lower = ++p;
  for (p = lower; *p != '.'; p++);
  *p = '\0';
  *chunk = strtol(lower, NULL, 10);
  
  free(upper);
}

void
parse_cg_filename(const char *name, int *id)
{
  char *str, *p, *q;

  str = strdup(name);

  if (!strncmp(name, "CGHANA", 6)) {
    *id = 2499;
    return;
  } else if (!strncmp(name, "CGBLACK", 7)) {
    *id = 2498;
    return;
  } else if (!strncmp(name, "HNTI01", 6)) {
    *id = 2497;
    return;
  } else if (!strncmp(name, "HNTI02", 6)) {
    *id = 2496;
    return;
  } else if (!strncmp(name, "HNTI03", 6)) {
    *id = 2495;
    return;
  } else if (!strncmp(name, "HNTI04", 6)) {
    *id = 2494;
    return;
  } else if (!strncmp(name, "HNTI05", 6)) {
    *id = 2493;
    return;
  }
  
  for (p = str; *p != '0'; p++);
  for (q = p; *q != '.'; q++);
  *q = '\0';

  *id = strtol(p, NULL, 10);

  free(str);
}

char **
make_filelist(const char *filename, int *num)
{
  FILE *fp;
  char buf[NAME_LEN];
  char fname[NAME_LEN];
  int i = 0;
  char **res;

  fprintf(stderr, "Reading file list[%s]\n", filename);

  fp = fopen(filename, "rb");
  if (fp == NULL) {
    fprintf(stderr, "Can't open outputfile %s.\n", filename);
    exit(1);
  }
  while (fgets(buf, NAME_LEN, fp) != NULL) {
    i++;
  }
  *num = i;
  fseek(fp, 0L, SEEK_SET);
  res = (char **)calloc(i, sizeof(char *));
  if (res == NULL) {
    perror("calloc");
    exit(1);
  }
  for (i = 0; i < *num; i++) {
    if (fgets(buf, NAME_LEN, fp) == NULL) {
      perror("fgets");
      exit(1);
    }
    sscanf(buf, "%s", fname);
    res[i] = strdup(fname);
    fprintf(stderr, "%s\n", res[i]);
  }
  fclose(fp);

  return res;
}

int
main(int argc, char *argv[])
{
  FILE *fp;
  int num, arg;
  fileinfo *fi;
  char **flist;

  uid = 0;

  if (argc < 3) {
    fprintf(stderr, "usage: pdbtar [-f filelist] archive file_1 ... file_N\n");
    exit(1);
  }
  if (!strncmp(argv[1], "-f", 2)) {
    arg = 4;
    flist = make_filelist(argv[2], &num);
  } else {
    arg = 2;
    flist = argv + 2;
    num = argc - arg;
  }

  fp = fopen(argv[arg - 1], "wb");
  if (fp == NULL) {
    fprintf(stderr, "Can't open outputfile %s.\n", argv[arg - 1]);
    exit(1);
  }

  fi = collect_file_info(num, flist);

  output_header(fp, "Hanashizume no Matsuri for Palm", "HTMK", num + 1);
  output_record_entry_header(fp, fi, num);
  output_delimiter(fp);
  output_app_info();
  output_sort_info();

  output_first_record(fp, fi, num);
  output_record(fp, fi, num);

  fclose(fp);
  free(fi);

  return 0;
}
